"""
Settings for configuring the application general behavior, plugging collection and other
aspects of the tool.
"""
from typing import List

from .plugins import collect_builtin_extensions

APP_LONG_NAME: str = "Python GUI Template"
"""Long, human meaningful app name."""

PLUGINS: List[str] = []
"""Explicit list of plugins that will be loaded in the order provided."""

AUTO_PLUGINS: List[str] = collect_builtin_extensions()
"""Plugins collected from the given locations. See `plugins.collect_plugins`"""

NOTEBOOK_LAYOUT: bool = True
"""Indicate if a notebook layout should be used in contrast to a central widget only."""

TAB_STYLE: str = "top"
"""Location of the tabs. Valid values are `top`, `bottom`, `left` and `right`."""



import wx

from pyguitemp.plugins import MenuTool, PluginBase


class MenuEntryPlugin(PluginBase):
    def menu_entries(self):
        save = MenuTool(
            menu="File",
            text="Save data",
            description="Save selected data into disk",
            callback=save_data,
        )
        load = MenuTool(
            menu="File",
            text="Load data",
            description="Load new data from disk",
            callback=load_data,
        )
        return [save, load]


def save_data(_):
    dial = wx.MessageDialog(
        None,
        "You click the 'Save data' menu entry",
        "Saving data...",
        wx.OK | wx.STAY_ON_TOP | wx.CENTRE,
    )
    dial.ShowModal()


def load_data(_):
    dial = wx.MessageDialog(
        None,
        "You click the 'Load data' menu entry",
        "Loading data...",
        wx.OK | wx.STAY_ON_TOP | wx.CENTRE,
    )
    dial.ShowModal()



"""
Contains all the custom machinery for dealing with logging in the program. It should
allow for new handlers to be added if any plugin requires it.
"""
import logging
import os
import sys
from datetime import datetime
from pathlib import Path
from typing import Optional

from . import APP_NAME


def app_dir(app_name: str = APP_NAME) -> Path:
    """Finds the application data directory for the current platform.

    If it does not exists, it creates the required directory tree.

    Returns:
        The path to the root app directory.
    """
    if sys.platform == "win32":
        path = Path.home() / "AppData" / "Local" / app_name
    elif sys.platform == "darwin":
        path = Path.home() / "Library" / "Application Support" / app_name
    else:
        path = Path.home() / f".{app_name}"

    _create_tree(path)
    return path


def _create_tree(path: Path) -> None:
    """Creates the app directory tree.

    Args:
        path: Where the directory tree must be created
    """
    os.makedirs(str(path), exist_ok=True)
    os.makedirs(str(path / "logs"), exist_ok=True)


class Logger:
    def __init__(self, app_name: str = APP_NAME):
        self._logger: Optional[logging.Logger] = None
        self.app_name = app_name

    @property
    def logger(self) -> logging.Logger:
        """Interface to the underlying logger.

        The logger is created automatically the first time it is used.

        Returns:
            A Logger object.
        """
        if not isinstance(self._logger, logging.Logger):
            self._logger = logging.getLogger("root")
            self._logger.setLevel(logging.DEBUG)

            self.set_console_handler()
            self.set_file_handler()

        return self._logger

    def __getattr__(self, item):
        return getattr(self.logger, item)

    def set_console_handler(self):
        """Sets a handler to print the log to the terminal."""
        ch = logging.StreamHandler()
        ch.setLevel(logging.DEBUG)

        # create and add formatter to handle
        formatter = logging.Formatter(
            fmt="%(asctime)s [%(levelname)s] %(module)s.%(funcName)s(%(lineno)d): %("
            "message)s",
            datefmt="%Y/%m/%d %H:%M:%S",
        )
        ch.setFormatter(formatter)

        # add ch to logger
        self._logger.addHandler(ch)

    def set_file_handler(self):
        """Sets a handler to print the log to a file in the app directory."""
        filename = (
            app_dir(self.app_name)
            / "logs"
            / f"{datetime.now().strftime('%Y%m%d_%H-%M-%S')}.log"
        )
        ch = logging.FileHandler(str(filename))
        ch.setLevel(logging.INFO)

        # create and add formatter to handle
        formatter = logging.Formatter(
            fmt="%(asctime)s [%(levelname)s]: %(message)s", datefmt="%Y/%m/%d %H:%M:%S"
        )
        ch.setFormatter(formatter)

        # add ch to logger
        self.logger.addHandler(ch)


logger = Logger()
"""Global logger for the application"""



import wx

from pyguitemp.plugins import PluginBase, Tab


class NotebookPlugin(PluginBase):
    def tabs(self, parent):
        text1 = Tab(
            page=wx.TextCtrl(parent, style=wx.TE_MULTILINE),
            text="Text area",
        )
        text2 = Tab(
            page=wx.TextCtrl(parent, style=wx.TE_MULTILINE),
            text="A second text area",
        )
        return [text1, text2]



def test_logger(caplog):
    from pyguitemp.logging import app_dir, logger

    msg = "Some function executed."

    def some_func():
        logger.info(msg)

    some_func()
    assert caplog.records[-1].levelname == "INFO"
    assert caplog.records[-1].message == msg

    log_dir = app_dir() / "logs"
    last_log = sorted(log_dir.glob("*.log"))[-1]
    with last_log.open() as f:
        log = list(f)

    assert "INFO" in log[-1]
    assert msg in log[-1]



from unittest.mock import MagicMock, patch

import pytest


@pytest.mark.xfail
def test_run_app():
    class App:
        MainLoop = MagicMock()

    with patch("pyguitemp.scripts.MainApp", MagicMock(return_value=App)):
        from pyguitemp.scripts import MainApp, run

        run()

        MainApp.assert_called_once()
        App.MainLoop.assert_called_once()


def test_init_repo(tmpdir):
    with patch("pyguitemp.scripts.copytree", MagicMock()):
        from pathlib import Path

        from pyguitemp.scripts import _init_repo, copytree

        skeleton = Path(__file__).parent.parent / "pyguitemp" / "skeleton"
        _init_repo(tmpdir, "my_app")
        copytree.assert_called_once_with(skeleton, tmpdir / "my_app")


class TestRunSubCommand:
    def test_add_arguments(self):
        import argparse

        from pyguitemp.scripts import RunSubCommand

        command = RunSubCommand()
        parser = argparse.ArgumentParser()
        command.add_arguments(parser)
        assert set([a.dest for a in parser._actions]) == {"help"}

    def test_run(self):
        import argparse

        with patch("pyguitemp.scripts.run", MagicMock()):
            from pyguitemp.scripts import RunSubCommand, run

            command = RunSubCommand()
            command.run(argparse.Namespace())
            run.assert_called_once()


class TestInitSubCommand:
    def test_add_arguments(self):
        import argparse

        from pyguitemp.scripts import InitSubCommand

        command = InitSubCommand()
        parser = argparse.ArgumentParser()
        command.add_arguments(parser)
        assert set([a.dest for a in parser._actions]) == {"help", "name", "target"}

    def test_run(self):
        import argparse

        with patch("pyguitemp.scripts._init_repo", MagicMock()):
            from pyguitemp.scripts import InitSubCommand, _init_repo

            command = InitSubCommand()
            args = argparse.Namespace(target="", name="")
            command.run(args)
            _init_repo.assert_called_once()


def test_parse_args():

    with patch("argparse.ArgumentParser.parse_args", MagicMock()):
        from pyguitemp.scripts import _SUB_COMMANDS, _parse_args

        for command in _SUB_COMMANDS:
            command.add_arguments = MagicMock()

        _parse_args()

        for command in _SUB_COMMANDS:
            command.add_arguments.assert_called_once()


def test_main():
    class Namespace:
        command = "run"

    with patch("pyguitemp.scripts._parse_args", MagicMock(return_value=Namespace)):
        from pyguitemp.scripts import _SUB_COMMAND_BY_NAME, _parse_args, main

        _SUB_COMMAND_BY_NAME[Namespace.command].run = MagicMock()

        main()

        _parse_args.assert_called_once()
        _SUB_COMMAND_BY_NAME[Namespace.command].run.assert_called_once()



"""
Contains all the machinery to register and load plugins.
"""
from __future__ import annotations

import importlib
import inspect
import sys
from abc import ABC
from dataclasses import dataclass
from pathlib import Path
from typing import Callable, List, Optional, Type

import wx

from .logging import logger

KNOWN_PLUGINS: List[Type[PluginBase]] = []
"""List of plugins registered as subclasses of PluginBase."""


@dataclass
class MenuTool:
    """Compiles the information required to add menu and toolbar items."""

    menu: str = ""
    id: int = wx.ID_ANY
    text: str = ""
    description: str = ""
    short_help: str = ""
    callback: Optional[Callable[[wx.Event], None]] = None
    kind: wx.ItemKind = wx.ITEM_NORMAL
    bitmap: Optional[wx.Bitmap] = None


@dataclass
class Tab:
    """Compiles the information required to define add tabs."""

    page: wx.Window
    text: str
    select: bool = False
    imageId: int = wx.NO_IMAGE
    order: int = 0


class PluginBase(ABC):
    """
    Base class that defines the required API that all the views of the different plugins
    will need to provide.
    """

    def __init_subclass__(cls: Type[PluginBase]):
        if cls not in KNOWN_PLUGINS:
            KNOWN_PLUGINS.append(cls)

    def menu_entries(self) -> List[MenuTool]:
        """Return a list of menu entries provided by this plugin.

        The items of the list must be MenuTool objects.

        Returns:
            A list of MenuTool objects provided by this plugin to be added to the
            application menu bar.
        """
        return []

    def toolbar_items(self) -> List[MenuTool]:
        """Return a list of toolbar items provided by this plugin.

        The items of the list must be MenuTool objects.

        Returns:
            A list of MenuTool objects provided by this plugin to be added to the
            application toolbar.
        """
        return []

    def tabs(self, parent=None) -> List[Tab]:
        """Create and return the list of tabs provided by this plugin.

        The elements of the list must be Tab objects.

        Args:
            parent: The application notebook, to be used as parent during the creation
            of the tabs.

        Returns:
            A list of Tab objects provided by this plugin to be added to the application
            notebook.
        """
        return []

    def central(self, parent=None) -> Optional[wx.Window]:
        """Central widget provided by this plugin.

        Args:
            parent: The main window, to be used as parent during the creation of the
            widget.

        Returns:
            None (if no central widget provided) or any object derived from
            wx.MainWindow (pretty much, any widget).
        """
        return None


def collect_plugins(
    path: Path, package: Optional[str] = None, add_to_path: bool = False
) -> List[str]:
    """Collects the plugins from the given location.

    This function collects the modules and packages available in chosen location. If
    they are not directly importable by Python (i.e. they are not in the path), either
    the name of the containing package should be supplied or the add_to_path flag must
    be set to True.

    Eg. If path = "/some/location/with/tools/mytool", and "mytool" is an importable
    package with "import mytool", plugins within "mytool" folder will be included with
    no extra arguments.

    If "mytool" is not importable but "tools" is, i.e. 'import tools.mytool' works,
    then you should add "package='tools'".

    If neither "mytool" nor anything else in its path is importable, then the path can
    be added to the module's import path, sys.path, so the plugins within can be
    imported as "import plugin".

    Args:
        path: Directory to explore.
        package: Package in which the directory is contained.
        add_to_path: If the directory should be added to the import path.

    Returns:
        List of modules and packages to be imported in the form
        "package.subpackage.plugin", where "plugin" can be a module (a python file) or
        a package itself (a folder containing a __init__.py file).
    """
    if add_to_path:
        sys.path.insert(1, str(path))
        pkg = ""
    else:
        pkg = f"{path.stem}." if package is None else f"{package}.{path.stem}."

    plugin_names = []
    for p in path.glob("*.py"):
        if not p.stem.startswith("__"):
            plugin_names.append(f"{pkg}{p.stem}")

    for p in path.glob("*/"):
        if p.is_dir() and (p / "__init__.py").exists():
            plugin_names.append(f"{pkg}{p.stem}")

    return plugin_names


def collect_builtin_extensions():
    """Search for plugins to be loaded in the "extensions" subfolder.

    These plugins can be single modules (simple_plugin.py) or more complex modules
    defined as packages (i.e. they are folders with a __init__.py file).

    Returns:
        A list of plugins names to be loaded.
    """
    frame = inspect.stack()[1]
    caller_file = Path(frame[0].f_code.co_filename)
    extensions = caller_file.parent / "extensions"
    return collect_plugins(extensions, Path(__file__).parent.stem)


def load_plugins(plugin_list: List[str]):
    """Loads the plugins of the list.

    Args:
        plugin_list: A list of plugins to be loaded.
    """
    for plugin in plugin_list:
        try:
            importlib.import_module(plugin)
        except ModuleNotFoundError as err:
            logger.warning(f"Plugin '{err.name}' could not be loaded. {err}")



from __future__ import annotations

import argparse
import importlib
import inspect
from abc import ABC, abstractmethod
from pathlib import Path
from shutil import copytree
from typing import Dict, List

from . import APP_NAME
from . import config as dconfig
from .core import MainApp
from .logging import logger
from .plugins import collect_builtin_extensions

logger.app_name = APP_NAME


def run():
    """Runs pyguitemp as an application, loading all the plugins available."""
    frame = inspect.stack()[1]
    caller_file = Path(frame[0].f_code.co_filename)
    config_file = caller_file.parent / "config.py"

    if not config_file.exists():
        raise RuntimeError(f"Configuration file '{config_file}' not found!")

    config = importlib.import_module(f"{caller_file.parent.stem}.config")

    title = config.APP_LONG_NAME if "APP_LONG_NAME" in dir(config) else "My App"
    plug = config.PLUGINS if "PLUGINS" in dir(config) else []
    autoplugins = (
        config.AUTO_PLUGINS if "AUTO_PLUGINS" in dir(config) else dconfig.AUTO_PLUGINS
    )
    nb_layout = (
        config.NOTEBOOK_LAYOUT
        if "NOTEBOOK_LAYOUT" in dir(config)
        else dconfig.NOTEBOOK_LAYOUT
    )
    tab_style = config.TAB_STYLE if "TAB_STYLE" in dir(config) else dconfig.TAB_STYLE

    app = MainApp(
        title=title,
        plugins_list=plug + autoplugins,
        notebook_layout=nb_layout,
        tab_style=tab_style,
    )
    app.MainLoop()


def _init_repo(path: Path, name: str):
    """Initialises repository with the files/folders required to use pyguitemp.

    This is done by copying the skeleton repository structure within pyguitemp to the
    desired location.

    Args:
        path: Where the repository should be initialized. Default: current.
        name: Name of the application.
    """
    full_dir = path / name
    logger.info(f"Initialising '{name}' in `{full_dir}`...")
    copytree(Path(__file__).parent / "skeleton", full_dir)


def _list_plugins() -> None:
    """Lists all plugins available in pyguitemp that the user can copy."""
    plug = [ext.split(".")[-1] for ext in collect_builtin_extensions()]
    for p in plug:
        print(p)


def _copy_plugin(name: List[str], target: Path) -> None:
    """Lists all plugins available in pyguitemp that the user can copy.

    Args:
        name: List of plugins to copy.
        target: Root destination for the plugins.
    """
    available = [ext.split(".")[-1] for ext in collect_builtin_extensions()]
    root = Path(__file__).parent / "extensions"

    for p in name:
        if p not in available:
            raise ValueError(
                f"Plugin '{p}' unknown. Run with '-l' flag for a list of valid values"
            )

        logger.info(f"Copying '{p}' into `{target}`...")
        copytree(root / p, target / p)


class SubCommand(ABC):
    def __init__(self, name, description):
        self.name = name
        self.description = description

    @abstractmethod
    def add_arguments(self, parser: argparse.ArgumentParser):
        pass

    @abstractmethod
    def run(self, args: argparse.Namespace):
        pass


class RunSubCommand(SubCommand):
    def __init__(self):
        super().__init__("run", f"Run '{APP_NAME}' as a standalone app.")

    def add_arguments(self, parser: argparse.ArgumentParser):
        pass

    def run(self, args: argparse.Namespace):
        run()


class InitSubCommand(SubCommand):
    def __init__(self):
        super().__init__(
            "init", "Initialises the current directory with an skeleton of a GUI app"
        )

    def add_arguments(self, parser: argparse.ArgumentParser):
        parser.add_argument(
            "name",
            type=str,
            help="Name of the app. To be used as package name.",
        )
        parser.add_argument(
            "-t",
            "--target",
            default=".",
            type=str,
            help="Target location for the app. Default: Current directory.",
        )

    def run(self, args: argparse.Namespace):
        _init_repo(Path(args.target).absolute(), args.name)


class PluginSubCommand(SubCommand):
    def __init__(self):
        super().__init__(
            "plugin", "Brings exisitng plugins in pyguitemp to the application."
        )

    def add_arguments(self, parser: argparse.ArgumentParser):
        parser.add_argument(
            "-l",
            "--list",
            action="store_true",
            help="Lists the plugins available.",
        )
        parser.add_argument(
            "-n",
            "--name",
            default="",
            type=str,
            help="Lists of comma-separated names of the plugins to copy.",
        )
        parser.add_argument(
            "-t",
            "--target",
            default=".",
            type=str,
            help="Target location for the plugin. Default: Current directory.",
        )

    def run(self, args: argparse.Namespace):
        if args.list:
            _list_plugins()
        elif args.name != "":
            name = args.name.split(",")
            target = Path(args.target).absolute()
            _copy_plugin(name, target)
        else:
            print("Nothing to do...")


_SUB_COMMANDS: List[SubCommand] = [
    RunSubCommand(),
    InitSubCommand(),
    PluginSubCommand(),
]

_SUB_COMMAND_BY_NAME: Dict[str, SubCommand] = {
    sub_command.name: sub_command for sub_command in _SUB_COMMANDS
}


def _parse_args(argv: List[str] = None) -> argparse.Namespace:
    """Parse the input arguments list populating the relevant subcommands

    Args:
        argv: List of arguments

    Returns:
        The namespace containing the relevant sub command information and arguments.
    """
    parser = argparse.ArgumentParser(prog=APP_NAME)
    subparsers = parser.add_subparsers(dest="command")
    subparsers.required = True
    for sub_command in _SUB_COMMANDS:
        sub_parser = subparsers.add_parser(
            sub_command.name, help=sub_command.description
        )
        sub_command.add_arguments(sub_parser)

    return parser.parse_args(argv)


def main(argv: List[str] = None):
    """Main entry point for pyguitemp, calling the relevant subcommands.

    Args:
        argv: List of input arguments, including subcommand.

    Returns:

    """
    args = _parse_args(argv)
    _SUB_COMMAND_BY_NAME[args.command].run(args)
